unit Unit3;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, Sockets, Menus;

type
  // This sample creates a socket using TAsyncSocket in TObject mode as
  // opposed to TComponent.  The procedures are the same basically for both
  // with the exception of using override statements instead of clicking and
  // obtaining the 'OnEvent' code in the designer.  You will still probably
  // need to write a new class for any project of considerable complexity even
  // if you use the TComponent version, so the TObject version is demonstrated
  // here.

  // Notice the EVENT_LISTEN in the Sockets unit.  It is a define that will
  // tell the code what events to listen for on a basic listening socket.
  // If you want the events changed, change the define and recompile.  Pay
  // close attention to these defines because no matter what socket objects
  // you inherit and how you override their functions, the initial events
  // will NOT be triggered unless they are defined in the EVENT area.  The only
  // exceptions to this are the OnListen and OnError.  They will be called
  // always because of the way they are implemented.  If you wish them to do
  // nothing, either override their respective procedures or remove the
  // message box defines.  (They only pop a message box up in their default
  // states).

   // TServerSocket - the socket that will listen for new connections.
  TServerSocket = class(TAsyncSocket)
  public
    m_Status: PChar;
    procedure OnListen(SocketMessage: TWMSocket); override;
    procedure OnAccept(SocketMessage: TWMSocket); override;
    procedure OnError(SocketMessage: TWMSocket); override;
    procedure OnClose(SocketMessagE: TWMSocket); override;
  end;

  // TClientSocket - the socket that will be attached to an accept request.
  TClientSocket = class(TAsyncSocket)
  public
    m_ListItem: TListItem;
    procedure OnRead(SocketMessage: TWMSocket); override;
    procedure OnError(SocketMessage: TWMSocket); override;
    procedure OnClose(SocketMessage: TWMSocket); override;
    procedure SendString(fString: String);
  end;

  TForm3 = class(TForm)
    ListView1: TListView;
    StatusBar1: TStatusBar;
    MainMenu1: TMainMenu;
    Server1: TMenuItem;
    Listen1: TMenuItem;
    PortNumber1: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Listen1Click(Sender: TObject);
    procedure PortNumber1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    m_ServerSocket: TServerSocket;
    m_GarbageList: TList;
    m_MessageList: TStringList;
    m_strBuff: Array[0..2047] of Char;
    procedure AddSocket(fSocket: TClientSocket);
    procedure OnIdle(Sender: TObject; var Done: Boolean);
    procedure AddSocketToGarbage(fSocket: TClientSocket);
    procedure BroadcastMessages;
  end;

var
  Form3: TForm3;

implementation

{$R *.DFM}

uses
  Unit1, Unit2, Unit4;  // This is to prevent a brutal circular reference in Pascal.
                        // if you put this in the upper uses clause you will get an
                        // error.  In my opinion, one of the BIG reasons to use
                        // Borland's C++ Builder.

// Note that these overrides do not need to call the base class because
// the base class procedures do nothing except put a message box on the screen.
// Read over TWMSocket to see the various data that can be extracted from it.
procedure TServerSocket.OnAccept(SocketMessage: TWMSocket);
var
  tempSocket: TClientSocket;
begin
  tempSocket := TClientSocket.Create;
  DoAccept(TAsyncSocket(tempSocket));
  // Now, assuming no errors, the tempSocket is initialized to the incoming
  // connection.  It is ready to be used.  We will now ship it off to the
  // TListView to have it added to the display.
  Form3.AddSocket(tempSocket);
  Form2.RichEdit1.Lines.Add('*SYSTEM* ACCEPT ON SERVERSOCKET ' +
    IntToStr(SocketMessage.SocketNumber));
end;

procedure TServerSocket.OnListen(SocketMessage: TWMSocket);
begin
  Form2.RichEdit1.Lines.Add('*SYSTEM* LISTEN ON SERVERSOCKET ' +
    IntToStr(SocketMessage.SocketNumber));
  m_Status := 'Listening';
end;

procedure TServerSocket.OnError(SocketMessage: TWMSocket);
begin
  Form2.RichEdit1.Lines.Add('*ERROR* ON SERVERSOCKET ' +
    IntToStr(SocketMessage.SocketNumber) + ' ' + ErrToStr(SocketMessage.SocketError));
end;

procedure TServerSocket.OnClose(SocketMessage: TWMSocket);
begin
  Form2.RichEdit1.Lines.Add('*SYSTEM* CLOSE ON SERVERSOCKET ' +
    IntToStr(SocketMessage.SocketNumber));
  m_Status := 'Idle';
end;

procedure TClientSocket.OnClose(SocketMessage: TWMSocket);
begin
  Form2.RichEdit1.Lines.Add('*SYSTEM* CLOSE ON CLIENTSOCKET ' +
    IntToStr(SocketMessage.SocketNumber));
  Form3.m_MessageList.Add('(ID:  ' + IntToStr(SocketMessage.SocketNumber) + ')  SIGNED OFF');
  Form3.AddSocketToGarbage(Self);
end;

procedure TClientSocket.OnRead(SocketMessage: TWMSocket);
var
  len: LongInt;
begin
  len := SocketMessage.SocketDataSize;
  if (len < 2048) then
    begin
    ZeroMemory(@Form3.m_strBuff, 2048);
    DoReceive(@Form3.m_strBuff, len);
    Form3.m_MessageList.Add('(ID:  ' + IntToStr(SocketHandle) + ')  ' + Form3.m_strBuff);
    end
  else
    Form4.RichEdit1.Lines.Add('Message too long to receive');
end;

procedure TClientSocket.OnError(SocketMessage: TWMSocket);
begin
  Form2.RichEdit1.Lines.Add('*ERROR* ON CLIENTSOCKET ' +
    IntToStr(SocketMessage.SocketNumber) + ' ' + ErrToStr(SocketMessage.SocketError));
    // ErrToStr converts a socket error into its descriptive string.  Note that
    // the TWMSocket has a collection of useful data items within it.  They
    // are only assured to be assigned within their respective functions.
    // TWMSocket.SocketNumber is filled in all instances.  TWMSocketNumber will
    // be -1 if the socket has not yet been allocated or has failed.  In most
    // instances, it will return the socket's number.
  DoClose;
end;

procedure TClientSocket.SendString(fString: String);
var
  tempLength: LongInt;
begin
  tempLength := Length(fString);
  DoSend(PChar(fString), tempLength);
end;

procedure TForm3.AddSocket(fSocket: TClientSocket);
var
  tempListItem: TListItem;
  ipadd: string;
begin
  tempListItem := ListView1.Items.Add;
  tempListItem.Data := fSocket; // This way we can find the socket if we need to
                                // from within the ListView.
  fSocket.m_ListItem := tempListItem; // If the socket is called by an event that
                                      // needs to update the ListView, we are
                                      // able to quickly identify the ListView
                                      // item from within the socket itself.
  ipadd := fSocket.DoGetHostByAddr(PChar(fSocket.IPAddress));
  if (ipadd = '') then
    ipadd := fSocket.IPAddress;
  tempListItem.Caption := 'Socket:  ' + IntToStr(fSocket.SocketHandle) + ' from ' + ipadd;
  // The above demonstrates how to utilize the DoGetHost functions in your
  // application.  You can provide the string, but in this case, we want the
  // location of the socket itself so we provide its own IPAddress.
end;

procedure TForm3.FormCreate(Sender: TObject);
begin
  m_GarbageList := TList.Create;
  m_MessageList := TStringList.Create;
  m_ServerSocket := TServerSocket.Create;
  m_ServerSocket.m_Status := 'Idle';
  m_ServerSocket.PortNumber := 2001;
  Application.OnIdle := OnIdle;
end;

procedure TForm3.FormDestroy(Sender: TObject);
begin
  m_GarbageList.Free;
end;

procedure TForm3.OnIdle(Sender: TObject; var Done: Boolean);
begin
// You should try to avoid freeing objects from within themselves, which is
// especially true of sockets.  Sockets can cause events and such after the
// fact so you are opening yourself up to a world of pain if you choose to
// free the socket from within its OnClose handler.  Further, this is the
// way that applications should update status bars and menus according to the
// fellows who wrote that foundation class thing.  They do it in an update
// handler - which is called during the idle time of the application.  See
// below for a good way to do this.

  while (m_GarbageList.Count > 0) do
    begin
    TClientSocket(m_GarbageList.First).m_ListItem.Free; // Free its ListItem
    TClientSocket(m_GarbageList.First).Free; // Free the socket itself
    m_GarbageList.Remove(m_GarbageList.First); // Remove it from the list
    end;

  if (m_MessageList.Count > c_ZERO) then
    BroadcastMessages;

  if (StatusBar1.Panels.Items[1].Text <> m_ServerSocket.m_Status) then
    StatusBar1.Panels.Items[1].Text := m_ServerSocket.m_Status;  // Prevents flickering
  if (StatusBar1.Panels.Items[0].Text <>
    ('Port Number:  ' + IntToStr(m_ServerSocket.PortNumber))) then
    StatusBar1.Panels.Items[0].Text := 'Port Number:  ' + IntToStr(m_ServerSocket.PortNumber);
  if (m_ServerSocket.m_Status = 'Listening') then
    begin
    if (Form3.Listen1.Checked = FALSE) then
      Form3.Listen1.Checked := TRUE;
    end
  else if (m_ServerSocket.m_Status = 'Idle') then
    begin
    if (Form3.Listen1.Checked = TRUE) then
      Form3.Listen1.Checked := FALSE;
    end;
  Done := TRUE;
end;

procedure TForm3.AddSocketToGarbage(fSocket: TClientSocket);
begin
  m_GarbageList.Add(fSocket);
end;

procedure TForm3.Listen1Click(Sender: TObject);
var
  status: LongInt;
begin
  if (Form3.m_ServerSocket.m_Status = 'Listening') then
    status := 1
  else
    status := 0;
  case status of
    0: // Idle
      Form3.m_ServerSocket.DoListen;
    1: // Listening
      Form3.m_ServerSocket.DoClose;
  end;
end;

procedure TForm3.PortNumber1Click(Sender: TObject);
begin
  try
    m_ServerSocket.PortNumber := StrToInt(InputBox('Port Number',
      'Input desired port number:', IntToStr(m_ServerSocket.PortNumber)));
  except
    Application.MessageBox('Port number must be 1-65535', 'Error', IDOK);
  end;
end;

procedure TForm3.BroadcastMessages;
var
  nStrings: LongInt;
  nSocketCount: LongInt;
begin
  nStrings := c_ZERO;
  while (nStrings < m_MessageList.Count) do
    begin
    nSocketCount := c_ZERO;
    while (nSocketCount < ListView1.Items.Count) do
      begin
      TClientSocket(ListView1.Items.Item[nSocketCount].Data).SendString(
        m_MessageList.Strings[nStrings]);
      Inc(nSocketCount);
      end;
    Form4.RichEdit1.Lines.Add(m_MessageList.Strings[nStrings]);
    Inc(nStrings);
    end;
  m_MessageList.Clear;
end;

end.
 